import ctypes
import os

import glm
from OpenGL.GL import *
from OpenGL.GL.shaders import *

os.environ["SDL_MAIN_USE_CALLBACKS"] = "1"
os.environ["SDL_RENDER_DRIVER"] = "opengl"

import sdl3
from panda3d.bullet import BulletWorld
from panda3d.core import Vec3

from box_drawer import BoxDrawer
from collider import Collider
from line_drawer import LineDrawer
from texture_loader import loadTexture

glContext = None
window = None

boxDrawer = None
crateTexture = None
green3Texture = None

lineDrawer = None

showCollider = True

canvasWidth = 400
canvasHeight = 400

# Create a perspective projection matrix and a view matrix
fovy = 45
near = 0.1
far = 100
projMatrix = glm.perspective(glm.radians(fovy),
    canvasWidth / canvasHeight, near, far)
viewMatrix = glm.lookAt(
    glm.vec3(10, 20, 25), # Camera position
    glm.vec3(0, 0, 0), # Target position
    glm.vec3(0, 1, 0)) # Camera up vector
# Combine them to one projView matrix
projViewMatrix = projMatrix * viewMatrix

world = BulletWorld()
world.setGravity(Vec3(0, -9.81, 0))

boxCollider = None
box2Collider = None
floorCollider = None

keys = { "left": False, "right": False, "up": False, "down": False }

def getCubeVertices(position, rotationQuat, halfSize):

    # Create a cube
    #    v5----- v1
    #   /|      /|
    #  v4------v0|
    #  | |     | |
    #  | |v7---|-|v3
    #  |/      |/
    #  v6------v2

    # Local vertices (untransformed, centered at origin)
    localVertices = [
        glm.vec3( halfSize,  halfSize,  halfSize),  # v0
        glm.vec3( halfSize,  halfSize, -halfSize),  # v1
        glm.vec3( halfSize, -halfSize,  halfSize),  # v2
        glm.vec3( halfSize, -halfSize, -halfSize),  # v3
        glm.vec3(-halfSize,  halfSize,  halfSize),  # v4
        glm.vec3(-halfSize,  halfSize, -halfSize),  # v5
        glm.vec3(-halfSize, -halfSize,  halfSize),  # v6
        glm.vec3(-halfSize, -halfSize, -halfSize)   # v7
    ]

    # Rotate and translate each vertex
    worldVertices = []
    for vertex in localVertices:
        # Apply rotation (quaternion * vertex)
        rotated = rotationQuat * vertex
        # Apply translation
        worldVertex = rotated + position
        worldVertices.append(worldVertex)

    return worldVertices

@sdl3.SDL_AppInit_func
def SDL_AppInit(appstate, argc, argv):
    global glContext
    global window
    global boxDrawer
    global crateTexture
    global green3Texture
    global boxCollider
    global box2Collider
    global floorCollider
    global lineDrawer

    if not sdl3.SDL_Init(sdl3.SDL_INIT_VIDEO):
        sdl3.SDL_Log("Couldn't initialize SDL: %s".encode() % sdl3.SDL_GetError())
        return sdl3.SDL_APP_FAILURE

    sdl3.SDL_GL_SetAttribute(sdl3.SDL_GL_MULTISAMPLEBUFFERS, 1) # Enable MULTISAMPLE
    sdl3.SDL_GL_SetAttribute(sdl3.SDL_GL_MULTISAMPLESAMPLES, 2) # Can be 2, 4, 8 or 16

    windowTitle = "PySDL3, PyGLM, Pillow, OpenGL, Python".encode()
    window = sdl3.SDL_CreateWindow(windowTitle, canvasWidth, canvasHeight,
        sdl3.SDL_WINDOW_OPENGL | sdl3.SDL_WINDOW_RESIZABLE)
    if not window:
        sdl3.SDL_Log("Couldn't create a window: %s".encode() % sdl3.SDL_GetError())
        return sdl3.SDL_APP_FAILURE

    # Create an OpenGL context
    glContext = sdl3.SDL_GL_CreateContext(window)
    if not glContext:
        sdl3.SDL_Log("Couldn't create a glContext: %s".encode() % sdl3.SDL_GetError())
        return sdl3.SDL_APP_FAILURE

    sdl3.SDL_GL_SetSwapInterval(1) # Turn on vertical sync

    glClearColor(0.15, 0.15, 0.15, 1)
    glEnable(GL_DEPTH_TEST)

    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    vertexShaderSource = None
    fragmentShaderSource = None

    with open("./assets/shaders/texture.vert") as file:
        vertexShaderSource = file.read()
    with open("./assets/shaders/texture.frag") as file:
        fragmentShaderSource = file.read()
    textureProgram = compileProgram(
       compileShader(vertexShaderSource, GL_VERTEX_SHADER),
       compileShader(fragmentShaderSource, GL_FRAGMENT_SHADER))

    with open("./assets/shaders/color.vert") as file:
        vertexShaderSource = file.read()
    with open("./assets/shaders/color.frag") as file:
        fragmentShaderSource = file.read()
    colorProgram = compileProgram(
       compileShader(vertexShaderSource, GL_VERTEX_SHADER),
       compileShader(fragmentShaderSource, GL_FRAGMENT_SHADER))

    boxDrawer = BoxDrawer(textureProgram)

    crateTexture = loadTexture("./assets/images/crate.png")
    green3Texture = loadTexture("./assets/images/green-3.png")

    floorName = "floor"
    floorPosition = glm.vec3(0, -10, 0)
    floorRotation = glm.quat(1, 0, 0, 0)
    floorSize = glm.vec3(20, 20, 20)
    floorCollider = Collider(floorName, floorPosition,
        floorRotation, floorSize, 0, world)
    floorCollider.node.setFriction(1)

    boxName = "box"
    boxPosition = glm.vec3(7, 2, 0)
    boxRotation = glm.quat(1, 0, 0, 0)
    boxSize = glm.vec3(3, 3, 3)
    boxCollider = Collider(boxName, boxPosition,
        boxRotation, boxSize, 10, world)
    boxCollider.node.setAngularFactor(Vec3(0, 0, 0))
    boxCollider.node.setFriction(1)
    boxCollider.node.setLinearSleepThreshold(False)

    box2Name = "box2"
    # box2Position = glm.vec3(2, 10, 0)
    box2Position = glm.vec3(0, 3, 0)
    box2Rotation = glm.quat(1, 0, 0, 0)
    box2Size = glm.vec3(3, 3, 3)
    box2Collider = Collider(box2Name, box2Position,
        box2Rotation, box2Size, 10, world)
    box2Collider.node.setFriction(1)
    box2Collider.node.setLinearSleepThreshold(False)

    # Create an object for drawing segments
    lineDrawer = LineDrawer(colorProgram, projViewMatrix)

    return sdl3.SDL_APP_CONTINUE

@sdl3.SDL_AppEvent_func
def SDL_AppEvent(appstate, event):
    global projViewMatrix

    if sdl3.SDL_DEREFERENCE(event).type == sdl3.SDL_EVENT_QUIT:
        return sdl3.SDL_APP_SUCCESS
    elif sdl3.SDL_DEREFERENCE(event).type == sdl3.SDL_EVENT_WINDOW_RESIZED:
        w = sdl3.SDL_GetWindowSurface(window).contents.w
        h = sdl3.SDL_GetWindowSurface(window).contents.h
        glViewport(0, 0, w, h)
        projMatrix = glm.perspective(glm.radians(fovy), w / h, near, far)
        projViewMatrix = projMatrix * viewMatrix
        lineDrawer.setProjViewMatrix(projViewMatrix)
    elif sdl3.SDL_DEREFERENCE(event).type == sdl3.SDL_EVENT_KEY_DOWN:
        if (
                sdl3.SDL_DEREFERENCE(event).key.scancode == sdl3.SDL_SCANCODE_W or
                sdl3.SDL_DEREFERENCE(event).button.button == sdl3.SDL_SCANCODE_UP
            ):
            keys["up"] = True
        if (
                sdl3.SDL_DEREFERENCE(event).key.scancode == sdl3.SDL_SCANCODE_S or
                sdl3.SDL_DEREFERENCE(event).button.button == sdl3.SDL_SCANCODE_DOWN
            ):
            keys["down"] = True
        if (
                sdl3.SDL_DEREFERENCE(event).key.scancode == sdl3.SDL_SCANCODE_A or
                sdl3.SDL_DEREFERENCE(event).button.button == sdl3.SDL_SCANCODE_LEFT
            ):
            keys["left"] = True
        if (
                sdl3.SDL_DEREFERENCE(event).key.scancode == sdl3.SDL_SCANCODE_D or
                sdl3.SDL_DEREFERENCE(event).button.button == sdl3.SDL_SCANCODE_RIGHT
            ):
            keys["right"] = True
    elif sdl3.SDL_DEREFERENCE(event).type == sdl3.SDL_EVENT_KEY_UP:
        if (
                sdl3.SDL_DEREFERENCE(event).key.scancode == sdl3.SDL_SCANCODE_W or
                sdl3.SDL_DEREFERENCE(event).button.button == sdl3.SDL_SCANCODE_UP
            ):
            keys["up"] = False
        if (
                sdl3.SDL_DEREFERENCE(event).key.scancode == sdl3.SDL_SCANCODE_S or
                sdl3.SDL_DEREFERENCE(event).button.button == sdl3.SDL_SCANCODE_DOWN
            ):
            keys["down"] = False
        if (
                sdl3.SDL_DEREFERENCE(event).key.scancode == sdl3.SDL_SCANCODE_A or
                sdl3.SDL_DEREFERENCE(event).button.button == sdl3.SDL_SCANCODE_LEFT
            ):
            keys["left"] = False
        if (
                sdl3.SDL_DEREFERENCE(event).key.scancode == sdl3.SDL_SCANCODE_D or
                sdl3.SDL_DEREFERENCE(event).button.button == sdl3.SDL_SCANCODE_RIGHT
            ):
            keys["right"] = False

    return sdl3.SDL_APP_CONTINUE

def keyboard():
    if keys["up"]:
        boxCollider.node.setLinearVelocity(Vec3(0, 0, -5))
    if keys["down"]:
        boxCollider.node.setLinearVelocity(Vec3(0, 0, 5))
    if keys["left"]:
        boxCollider.node.setLinearVelocity(Vec3(-5, 0, 0))
    if keys["right"]:
        boxCollider.node.setLinearVelocity(Vec3(5, 0, 0))

@sdl3.SDL_AppIterate_func
def SDL_AppIterate(appstate):

    keyboard()

    world.doPhysics(0.016)

    boxCollider.update()
    box2Collider.update()

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    if showCollider:
        vertices = getCubeVertices(box2Collider.position, box2Collider.rotation, 1.5)

        # Create a cube
        #    v5----- v1
        #   /|      /|
        #  v4------v0|
        #  | |     | |
        #  | |v7---|-|v3
        #  |/      |/
        #  v6------v2

        # v0 - v1
        lineDrawer.draw(start = vertices[0], end = vertices[1],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v0 - v2
        lineDrawer.draw(start = vertices[0], end = vertices[2],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v2 - v3
        lineDrawer.draw(start = vertices[2], end = vertices[3],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v3 - v1
        lineDrawer.draw(start = vertices[3], end = vertices[1],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v1 - v0
        lineDrawer.draw(start = vertices[1], end = vertices[0],
            color = glm.vec3(0, 1, 0), thickness = 0.2)

        # v4 - v5
        lineDrawer.draw(start = vertices[4], end = vertices[5],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v5 - v7
        lineDrawer.draw(start = vertices[5], end = vertices[7],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v7 - v6
        lineDrawer.draw(start = vertices[7], end = vertices[6],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v6 - v4
        lineDrawer.draw(start = vertices[6], end = vertices[4],
            color = glm.vec3(0, 1, 0), thickness = 0.2)

        # v0 - v4
        lineDrawer.draw(start = vertices[0], end = vertices[4],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v2 - v6
        lineDrawer.draw(start = vertices[2], end = vertices[6],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v3 - v7
        lineDrawer.draw(start = vertices[3], end = vertices[7],
            color = glm.vec3(0, 1, 0), thickness = 0.2)
        # v1 - v5
        lineDrawer.draw(start = vertices[1], end = vertices[5],
            color = glm.vec3(0, 1, 0), thickness = 0.2)

    axesSize = 50
    # X
    lineDrawer.draw(start = glm.vec3(-axesSize, 0, 0), end = glm.vec3(axesSize, 0, 0),
        color = glm.vec3(1, 0, 0), thickness = 0.1)
    # Y
    lineDrawer.draw(start = glm.vec3(0, -axesSize, 0), end = glm.vec3(0, axesSize, 0),
        color = glm.vec3(0, 1, 0), thickness = 0.1)
    # Z
    lineDrawer.draw(start = glm.vec3(0, 0, -axesSize), end = glm.vec3(0, 0, axesSize),
        color = glm.vec3(0, 0, 1), thickness = 0.1)

    # Static collider
    position = floorCollider.position
    size = floorCollider.size
    rotation = floorCollider.rotation
    boxDrawer.draw(projViewMatrix, position, size, rotation, green3Texture)

    # Box
    position = boxCollider.position
    size = boxCollider.size
    rotation = boxCollider.rotation
    boxDrawer.draw(projViewMatrix, position, size, rotation, crateTexture)

    # Box2
    position = box2Collider.position
    size = box2Collider.size
    rotation = box2Collider.rotation
    boxDrawer.draw(projViewMatrix, position, size, rotation, crateTexture)

    sdl3.SDL_GL_SwapWindow(window)
    return sdl3.SDL_APP_CONTINUE

@sdl3.SDL_AppQuit_func
def SDL_AppQuit(appstate, result):
    global glContext
    sdl3.SDL_GL_DestroyContext(glContext)
    # SDL will clean up the window/renderer for us
